﻿// Copyright (C) Microsoft Corporation. All Rights Reserved.
// This code released under the terms of the Microsoft Public License
// (Ms-PL, http://opensource.org/licenses/ms-pl.html).

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.IO;
using System.Linq;
using System.Resources;
using Microsoft.FxCop.Sdk;

// Settings for debugging Delay.FxCop within Visual Studio:
// C:\Program Files (x86)\Microsoft Visual Studio 10.0\Team Tools\Static Analysis Tools\FxCop\FxCopCmd.exe
// /file:..\TestProjects\ConsoleApplication\bin\Debug\ConsoleApplication.exe /ruleset:=..\Delay.FxCop.ruleset /console

namespace Delay.FxCop
{
    /// <summary>
    /// Customization of BaseIntrospectionRule for rules in the Delay.FxCop assembly.
    /// </summary>
    abstract class CustomBaseIntrospectionRule : BaseIntrospectionRule
    {
        /// <summary>
        /// Initializes a new instance of the CustomBaseIntrospectionRule class.
        /// </summary>
        /// <param name="ruleName">Name of the rule.</param>
        [SuppressMessage("Spelling", "DF1000:CheckSpellingOfAllStringLiterals", MessageId = "Fx", Justification = "Part of product name.")]
        protected CustomBaseIntrospectionRule(string ruleName)
            : base(ruleName, "Delay.FxCop.RuleMetadata", typeof(CustomBaseIntrospectionRule).Assembly)
        {
        }
    }

    /// <summary>
    /// Custom FxCop rule that checks the spelling of all string literals.
    /// </summary>
    class CheckSpellingOfAllStringLiterals : CustomBaseIntrospectionRule
    {
        /// <summary>
        /// Initializes a new instance of the CheckSpellingOfAllStringLiterals class.
        /// </summary>
        [SuppressMessage("Performance", "DF1002:UncalledMethodsShouldBeRemoved", Justification = "Instantiated by code analysis framework.")]
        public CheckSpellingOfAllStringLiterals()
            : base("CheckSpellingOfAllStringLiterals")
        { }

        /// <summary>
        /// Called to check a Member instance.
        /// </summary>
        /// <param name="member">Member instance.</param>
        /// <returns>Problem collection.</returns>
        public override ProblemCollection Check(Member member)
        {
            Visit(member);
            return Problems;
        }

        /// <summary>
        /// Called when analysis encounters a literal.
        /// </summary>
        /// <param name="literal">Literal instance.</param>
        public override void VisitLiteral(Literal literal)
        {
            if (null == literal)
            {
                throw new ArgumentNullException("literal");
            }
            if (FrameworkTypes.String == literal.Type)
            {
                CheckSpellingLiteral(literal);
            }
            base.VisitLiteral(literal);
        }

        /// <summary>
        /// Called when analysis encounters a field.
        /// </summary>
        /// <param name="field">Field instance.</param>
        public override void VisitField(Field field)
        {
            if (null == field)
            {
                throw new ArgumentNullException("field");
            }
            if (FrameworkTypes.String == field.Type)
            {
                CheckSpellingLiteral(field.DefaultValue);
            }
            base.VisitField(field);
        }

        /// <summary>
        /// Checks the spelling of the words for a literal instance.
        /// </summary>
        /// <param name="literal">Literal instance.</param>
        [SuppressMessage("Microsoft.Globalization", "CA1308:NormalizeStringsToUppercase", Justification = "Using lower-case to avoid confusion with acronyms.")]
        private void CheckSpellingLiteral(Literal literal)
        {
            if ((null != literal) && (null != literal.Value))
            {
                var namingService = NamingService.DefaultNamingService;
                foreach (var word in WordParser.Parse((string)literal.Value, WordParserOptions.IgnoreMnemonicsIndicators | WordParserOptions.SplitCompoundWords).Distinct())
                {
                    if ((WordSpelling.SpelledCorrectly != namingService.CheckSpelling(word)) &&
                        (WordSpelling.SpelledCorrectly != namingService.CheckSpelling(word.ToLowerInvariant())))
                    {
                        Problems.Add(new Problem(GetResolution(word), literal, word));
                    }
                }
            }
        }
    }

    /// <summary>
    /// Custom FxCop rule that verifies all resources are referenced.
    /// </summary>
    class ResourcesShouldBeReferenced : CustomBaseIntrospectionRule
    {
        /// <summary>
        /// Initializes a new instance of the ResourcesShouldBeReferenced class.
        /// </summary>
        [SuppressMessage("Performance", "DF1002:UncalledMethodsShouldBeRemoved", Justification = "Instantiated by code analysis framework.")]
        public ResourcesShouldBeReferenced()
            : base("ResourcesShouldBeReferenced")
        { }

        /// <summary>
        /// Called to check a Resource instance.
        /// </summary>
        /// <param name="resource">Resource instance.</param>
        /// <returns>Problem collection.</returns>
        [SuppressMessage("Microsoft.Usage", "CA2202:Do not dispose objects multiple times", Justification = "Safe and more explicit this way.")]
        public override ProblemCollection Check(Resource resource)
        {
            // Process the stream as a resource file
            using (var stream = new MemoryStream(resource.Data))
            {
                try
                {
                    using (var reader = new ResourceReader(stream))
                    {
                        // Check each item against the list of known references
                        foreach (var resourceName in
                            reader
                                .Cast<DictionaryEntry>()
                                .Select(e => e.Key.ToString())
                                .Where(n => !AssemblyResourceNameHelper.Get(resource.DefiningModule.ContainingAssembly).Contains(n)))
                        {
                            // Not referenced; report a problem
                            Problems.Add(new Problem(GetResolution(resourceName, resource.Name), resourceName));
                        }
                    }
                }
                catch (ArgumentException)
                {
                    // "Stream is not a valid resource file."
                }
            }
            return this.Problems;
        }
    }

    /// <summary>
    /// Custom FxCop rule that identifies uncalled methods of any kind.
    /// </summary>
    class UncalledMethodsShouldBeRemoved : CustomBaseIntrospectionRule
    {
        /// <summary>
        /// Initializes a new instance of the UncalledMethodsShouldBeRemoved class.
        /// </summary>
        [SuppressMessage("Performance", "DF1002:UncalledMethodsShouldBeRemoved", Justification = "Instantiated by code analysis framework.")]
        public UncalledMethodsShouldBeRemoved()
            : base("UncalledMethodsShouldBeRemoved")
        {
        }

        /// <summary>
        /// Called to check a Member instance.
        /// </summary>
        /// <param name="member">Member instance.</param>
        /// <returns>Problem collection.</returns>
        public override ProblemCollection Check(Member member)
        {
            Visit(member);
            return Problems;
        }

        /// <summary>
        /// Called when analysis encounters a Method.
        /// </summary>
        /// <param name="memberBinding">Current Method.</param>
        public override void VisitMethod(Method method)
        {
            base.VisitMethod(method);

            var declaringModule = method.DeclaringType.DeclaringModule;
            if ((null == method.OverriddenMethod) &&
                (null == method.ImplementedInterfaceMethods) &&
                !(method is StaticInitializer) &&
                (method != declaringModule.EntryPoint) &&
                !CallGraph.CallersFor(method).Any() &&
                !method.DeclaringType.Interfaces.Cast<InterfaceNode>().Where(n => null != n.GetMatchingMethod(method)).Any() &&
                !AssemblyReferencedMethodHelper.Get(declaringModule.ContainingAssembly).Contains(method))
            {
                Problems.Add(new Problem(GetResolution(method.FullName), method));
            }
        }
    }

    /// <summary>
    /// Class that exposes a sequence of resource names referenced by an AssemblyNode.
    /// </summary>
    class AssemblyResourceNameHelper : AssemblyNodeEnumerationHelper<string>
    {
        /// <summary>
        /// Gets a AssemblyResourceNameHelper instance for the specified AssemblyNode.
        /// </summary>
        /// <param name="assemblyNode">AssemblyNode.</param>
        /// <returns>AssemblyResourceNameHelper instance</returns>
        public static AssemblyResourceNameHelper Get(AssemblyNode assemblyNode)
        {
            return Get(assemblyNode, n => new AssemblyResourceNameHelper(n));
        }

        /// <summary>
        /// Initializes a new instance of the AssemblyResourceNameHelper class.
        /// </summary>
        /// <param name="assemblyNode">AssemblyNode of which to enumerate resource strings.</param>
        protected AssemblyResourceNameHelper(AssemblyNode assemblyNode)
            : base(assemblyNode)
        {
        }

        /// <summary>
        /// Called when analysis encounters a MethodCall.
        /// </summary>
        /// <param name="call">Current MethodCall.</param>
        /// <param name="caller">Current Method.</param>
        [SuppressMessage("Microsoft.Design", "CA1062:Validate arguments of public methods", MessageId = "0", Justification = "Only called by code analysis tool.")]
        [SuppressMessage("Microsoft.Design", "CA1062:Validate arguments of public methods", MessageId = "1", Justification = "Only called by code analysis tool.")]
        protected override void VisitMethodCall(MethodCall call, Method currentMethod)
        {
            base.VisitMethodCall(call, currentMethod);

            var callee = call.Callee as MemberBinding;
            if (null != callee)
            {
                var boundMethod = callee.BoundMember as Method;
                if (null != boundMethod)
                {
                    if ((boundMethod.DeclaringType.IsAssignableTo(FrameworkTypes.ResourceManager)) &&
                        (("GetString" == boundMethod.Name.Name) ||
                         ("GetStream" == boundMethod.Name.Name) ||
                         ("GetObject" == boundMethod.Name.Name)))
                    {
                        // Call to ResourceManager.GetString/GetStream/GetObject
                        if (CallGraph.CallersFor(currentMethod).Any() ||
                            !currentMethod.DeclaringType.Attributes.Where(a => ("System.CodeDom.Compiler.GeneratedCodeAttribute" == a.Type.FullName) && ("System.Resources.Tools.StronglyTypedResourceBuilder" == (string)((Literal)a.GetPositionalArgument(0)).Value)).Any())
                        {
                            // At least one caller of the Resources.Property OR a direct call to ResourceManager.GetXxx by something other than the generated Resources class
                            var literal = call.Operands.FirstOrDefault() as Literal;
                            if ((null != literal) && (FrameworkTypes.String == literal.Type))
                            {
                                Items.Add((string)literal.Value);
                            }
                        }
                    }
                }
            }
        }
    }

    /// <summary>
    /// Class that exposes a sequence of methods referenced by an AssemblyNode.
    /// </summary>
    class AssemblyReferencedMethodHelper : AssemblyNodeEnumerationHelper<Method>
    {
        /// <summary>
        /// Gets a AssemblyReferencedMethodHelper instance for the specified AssemblyNode.
        /// </summary>
        /// <param name="assemblyNode">AssemblyNode.</param>
        /// <returns>AssemblyReferencedMethodHelper instance</returns>
        public static AssemblyReferencedMethodHelper Get(AssemblyNode assemblyNode)
        {
            return Get(assemblyNode, n => new AssemblyReferencedMethodHelper(n));
        }

        /// <summary>
        /// Initializes a new instance of the AssemblyReferencedMethodHelper class.
        /// </summary>
        /// <param name="assemblyNode">AssemblyNode of which to enumerate referenced methods.</param>
        protected AssemblyReferencedMethodHelper(AssemblyNode assemblyNode)
            : base(assemblyNode)
        {
        }

        /// <summary>
        /// Called when analysis encounters a Parameter.
        /// </summary>
        /// <param name="parameter">Parameter.</param>
        [SuppressMessage("Microsoft.Design", "CA1062:Validate arguments of public methods", MessageId = "0", Justification = "Only called by code analysis tool.")]
        public override void VisitParameter(Parameter parameter)
        {
            base.VisitParameter(parameter);
            Items.Add(parameter.DeclaringMethod);
        }

        /// <summary>
        /// Called when analysis encounters a Construct.
        /// </summary>
        /// <param name="parameter">Construct.</param>
        [SuppressMessage("Microsoft.Design", "CA1062:Validate arguments of public methods", MessageId = "0", Justification = "Only called by code analysis tool.")]
        public override void VisitConstruct(Construct construct)
        {
            base.VisitConstruct(construct);
            foreach (var method in
                construct
                    .Operands
                    .OfType<UnaryExpression>()
                    .Select(ue => ue.Operand)
                    .OfType<MemberBinding>()
                    .Select(mb => mb.BoundMember)
                    .OfType<Method>())
            {
                Items.Add(method);
            }
        }
    }

    /// <summary>
    /// Class that exposes a sequence of items used or referenced by an AssemblyNode.
    /// </summary>
    class AssemblyNodeEnumerationHelper<T> : BinaryReadOnlyVisitor, IEnumerable<T>
    {
        /// <summary>
        /// Stores a cache of AssemblyNode to AssemblyNodeEnumerationHelper(T).
        /// </summary>
        private static Dictionary<AssemblyNode, AssemblyNodeEnumerationHelper<T>> _cache = new Dictionary<AssemblyNode, AssemblyNodeEnumerationHelper<T>>();

        /// <summary>
        /// Gets a AssemblyNodeEnumerationHelper(T) for the specified AssemblyNode.
        /// </summary>
        /// <typeparam name="H">Type of AssemblyNodeEnumerationHelper(T).</typeparam>
        /// <param name="assemblyNode">AssemblyNode.</param>
        /// <param name="constructor">Function that creates a AssemblyNodeEnumerationHelper(T) from an AssemblyNode.</param>
        /// <returns>AssemblyNodeEnumerationHelper(T) instance.</returns>
        protected static H Get<H>(AssemblyNode assemblyNode, Func<AssemblyNode, H> constructor) where H : AssemblyNodeEnumerationHelper<T>
        {
            lock (_cache)
            {
                if (!_cache.ContainsKey(assemblyNode))
                {
                    _cache.Add(assemblyNode, constructor(assemblyNode));
                }
            }
            return (H)_cache[assemblyNode];
        }

        /// <summary>
        /// Stores a reference to the AssemblyNode.
        /// </summary>
        private readonly AssemblyNode _assemblyNode;

        /// <summary>
        /// Stores a reference to the current Method during a Visit.
        /// </summary>
        private Method _currentMethod;

        /// <summary>
        /// Gets the list of items.
        /// </summary>
        protected List<T> Items { get; private set; }

        /// <summary>
        /// Initializes a new instance of the AssemblyNodeEnumerationHelper class.
        /// </summary>
        /// <param name="assemblyNode">AssemblyNode of which to enumerate items.</param>
        protected AssemblyNodeEnumerationHelper(AssemblyNode assemblyNode)
        {
            _assemblyNode = assemblyNode;
        }

        /// <summary>
        /// Called when analysis encounters a Method.
        /// </summary>
        /// <param name="memberBinding">Current Method.</param>
        public override void VisitMethod(Method method)
        {
            _currentMethod = method;
            base.VisitMethod(method);
        }

        /// <summary>
        /// Called when analysis encounters a MethodCall.
        /// </summary>
        /// <param name="call">Current MethodCall.</param>
        [SuppressMessage("Microsoft.Design", "CA1062:Validate arguments of public methods", MessageId = "0", Justification = "Only called by code analysis tool.")]
        public override void VisitMethodCall(MethodCall call)
        {
            base.VisitMethodCall(call);
            VisitMethodCall(call, _currentMethod);
        }

        /// <summary>
        /// Called when analysis encounters a MethodCall.
        /// </summary>
        /// <param name="call">Current MethodCall.</param>
        /// <param name="caller">Current Method.</param>
        protected virtual void VisitMethodCall(MethodCall call, Method currentMethod) { }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>An IEnumerator(T) that can be used to iterate through the collection.</returns>
        public IEnumerator<T> GetEnumerator()
        {
            lock (_assemblyNode) // Any non-null private member variable
            {
                if (null == Items)
                {
                    // Generate the list of items
                    Items = new List<T>();
                    Visit(_assemblyNode);
                }
            }

            // Return the enumerator
            return Items.AsReadOnly().GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>An IEnumerator object that can be used to iterate through the collection.</returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }
    }
}
